
/******************************************************************************
* MODULE     : scheme.gen.cc
* DESCRIPTION: fixed size schemes with reference counting
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <scheme.gen.h>

#module code_scheme
#import scheme

string TUPLE ("()");
string COLLECTION ("{,}");
string ASSOCIATE ("->");

/******************************************************************************
* Converting scheme trees to strings
******************************************************************************/

static string
slash (string s) {
  int i;
  string r;
  for (i=0; i<N(s); i++)
    switch (s[i]) {
    case '(':
    case ')':
    case ' ':
    case '\'':
    case '\\':
      r << "\\" << s[i];
      break;
    case '\"': // "
      if (((i==0) && (s[N(s)-1]=='\"')) || // "
	  ((i==(N(s)-1)) && (s[0]=='\"'))) // "
	r << s[i];
      else r << "\\" << s[i];
      break;
    case ((char) 0):
      r << "\\0";
      break;
    case '\t':
      r << "\\t";
      break;
    case '\n':
      r << "\\n";
      break;
    default:
      r << s[i];
    }
  return r;
}

static void
scheme_tree_to_string (string& out, scheme_tree p) {
  if (!is_tuple (p)) out << slash (p->label);
  else {
    int i;
    if (is_tuple (p, "\'", 1)) {
      out << "\'";
      scheme_tree_to_string (out, p[1]);
    }
    else {
      out << "(";
      for (i=0; i<N(p); i++) {
	if (i>0) out << " ";
	scheme_tree_to_string (out, p[i]);
      }
      out << ")";
    }
  }
}

string
scheme_tree_to_string (scheme_tree p) {
  string out;
  scheme_tree_to_string (out, p);
  return out;
}

/******************************************************************************
* Converting strings to scheme trees
******************************************************************************/

static bool
is_spc (char c) {
  return (c==' ') || (c=='\t') || (c=='\n');
}

static string
unslash (string s, int start, int end) {
  int i;
  string r;
  for (i=start; i<end; i++)
    if ((s[i]=='\\') && ((i+1)<end))
      switch (s[++i]) {
      case '0': r << ((char) 0); break;
      case 'n': r << '\n'; break;
      case 't': r << '\t'; break;
      default: r << s[i];
      }
    else r << s[i];
  return r;
}

static scheme_tree
string_to_scheme_tree (string s, int& i) {
  for (; i<N(s); i++)
    switch (s[i]) {

    case ' ':
    case '\t':
    case '\n':
      break;

    case '(':
      {
	scheme_tree p (TUPLE);
	i++;
	while (TRUE) {
	  while ((i<N(s)) && is_spc(s[i])) i++;
	  if ((i==N(s)) || (s[i]==')')) break;
	  p << string_to_scheme_tree (s, i);
	}
	if (i<N(s)) i++;
	return p;
      }

    case '\'':
      i++;
      return scheme_tree (TUPLE, "\'", string_to_scheme_tree (s, i));
      
    case '\"':
      { // "
	int start= i++;
	while ((i<N(s)) && (s[i]!='\"')) { // "
	  if ((i<N(s)-1) && (s[i]=='\\')) i++;
	  i++;
	}
	if (i<N(s)) i++;
	return scheme_tree (unslash (s, start, i));
      }

    case ';':
      while ((i<N(s)) && (s[i]!='\n')) i++;
      break;

    default:
      {
	int start= i;
	while ((i<N(s)) && (!is_spc(s[i])) && (s[i]!='(') && (s[i]!=')')) {
	  if ((i<N(s)-1) && (s[i]=='\\')) i++;
	  i++;
	}
	return scheme_tree (unslash (s, start, i));
      }
    }

  return "";
}

scheme_tree
string_to_scheme_tree (string s) {
  int i=0;
  return string_to_scheme_tree (s, i);
}

scheme_tree
block_to_scheme_tree (string s) {
  scheme_tree p (TUPLE);
  int i=0;
  while ((i<N(s)) && (is_spc (s[i]) || s[i]==')')) i++;
  while (i<N(s)) {
    p << string_to_scheme_tree (s, i);
    while ((i<N(s)) && (is_spc (s[i]) || s[i]==')')) i++;
  }
  return p;
}

/******************************************************************************
* Conversion between trees and scheme_trees
******************************************************************************/

scheme_tree
tree_to_scheme_tree (tree t) {
  int i, n=N(t);
  if (n==0) return "\"" * t->label * "\"";
  else {
    tree u (TUPLE, n+1);
    u[0]= copy (t->label);
    for (i=0; i<n; i++)
      u[i+1]= tree_to_scheme_tree (t[i]);
    return u;
  }
}

tree
scheme_tree_to_tree (scheme_tree t) {
  int i, n= N(t);
  if (n==0) {
    string& s= t->label;
    if ((N(s)>=2) && (s[0]=='\42') && (s[N(s)-1]=='\42'))
      return s (1, N(s)-1);
    else return s;
  }
  else {
    tree u (t[0]->label, n-1);
    for (i=1; i<n; i++)
      u[i-1]= scheme_tree_to_tree (t[i]);
    return u;
  }
}

#endmodule // code_scheme
