
/******************************************************************************
* MODULE     : tex_font.gen.cc
* DESCRIPTION: TeX text fonts
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <font.gen.h>
#include <tex.gen.h>

#module code_tex_font
#import font
#import tex

#define TEX_ANY   0
#define TEX_CM    1
#define TEX_EC    2
#define TEX_LA    3
#define TEX_ADOBE 4

/******************************************************************************
* TeX text fonts
******************************************************************************/

struct tex_font_rep: font_rep {
  int              status;
  tex_font_metric  tfm;
  bitmap_font      pk;
  double           unit;

  tex_font_rep (display dis, string name, int status,
		string family, int size, int dpi, int dsize);

  void get_extents (string s, text_extents& ex);
  void draw (ps_device dev, string s, SI x, SI y);
  bitmap_char get_bitmap (string s);
  void special_get_extents (string s, text_extents& ex);
  void special_draw (ps_device dev, string s, SI x, SI y);
  void accented_get_extents (string s, text_extents& ex);
  void accented_draw (ps_device dev, string s, SI x, SI y);
};

/******************************************************************************
* The implementation of tex_fonts
******************************************************************************/

#define conv(x) ((SI) (((double) (x))*unit))

tex_font_rep::tex_font_rep (display dis, string name, int status2,
  string family, int size, int dpi, int dsize):
    font_rep (dis, name), status (status2)
{
  load_tex (family, size, dpi, dsize, tfm, pk);

  design_size = tfm->design_size () >> 12;
  display_size= (((design_size*dpi)/72)*PIXEL) >> 8;
  unit        = ((double) display_size) / ((double) (1<<20));
  slope       = tfm->slope ();
  spc->def    = conv (tfm->spc ());
  spc->min    = spc->def - conv (tfm->spc_shrink ());
  spc->max    = spc->def + conv (tfm->spc_stretch ());
  extra       = conv (tfm->spc_extra ());
  extra->min  = extra->min >> 1;
  extra->max  = extra->min << 1;

  y1    = conv (-262080);
  y2    = y1+ display_size;
  ysub  = 0;
  ysup  = conv (tfm->x_height ());
  yfrac = (ysub + ysup) >> 1;
  wpt   = (dpi*PIXEL)/72;
  wquad = (wpt*design_size) >> 8;
  wunit = wquad/4;
  wfrac = (dpi<450)? (wquad/25): ((dpi<900)? (wquad/30): (wquad/35));
  wsqrt = (dpi<450)? (wquad/25): ((dpi<700)? (wquad/25): (wquad/30));
  wneg  = (dpi<450)? (wquad/25): ((dpi<700)? (wquad/25): (wquad/30));

  if (dpi<350) yfrac -= wquad/40;
  else if (dpi<500);
  else if (dpi<700) yfrac += wquad/80;
  else yfrac += wquad/60;
}

void
tex_font_rep::get_extents (string s, text_extents& ex) {
  register int i;
  switch (status) {
  case TEX_ANY:
    break;
  case TEX_CM:
    for (i=0; i<N(s); i++)
      if ((s[i] & 128) != 0) {
	accented_get_extents (s, ex);
	return;
      }
    break;
  case TEX_EC:
  case TEX_LA:
  case TEX_ADOBE:
    for (i=0; i<N(s); i++)
      if (s[i]=='<') {
	special_get_extents (s, ex);
	return;
      }
    break;
  }

  int n=N(s);
  int m=n+16;
  int s_copy[n], buf[m], ker[m];
  for (i=0; i<n; i++) s_copy[i]= ((QN) s[i]);
  tfm->execute (s_copy, n, buf, ker, m);

  SI x1= 0;
  SI x2= 0;
  SI x3= PLUS_INFINITY;
  SI x4= MINUS_INFINITY;
  SI y1= font_rep::y1;
  SI y2= font_rep::y2;
  SI y3= PLUS_INFINITY;
  SI y4= MINUS_INFINITY;

  for (i=0; i<m; i++) {
    int c= buf[i];
    bitmap_char bmc= pk->get (c);
    if (nil (bmc)) continue;
    
    x3= min (x3, x2- ((int) bmc->xoff) * PIXEL);
    x4= max (x4, x2+ ((int) (bmc->width- bmc->xoff)) * PIXEL);
    y3= min (y3, ((int) (bmc->yoff- bmc->height)) * PIXEL);
    y4= max (y4, ((int) bmc->yoff) * PIXEL);
    
    x2 += conv (tfm->w(c)+ ker[i]);
  }

  if ((x3 == PLUS_INFINITY) || (x4 == MINUS_INFINITY) ||
      (y3 == PLUS_INFINITY) || (y4 == MINUS_INFINITY))
    {
      x1=x3= x2=x4=0;
      y3=y1; y4=y2;
    }

  ex->x1= x1;
  ex->x2= x2;
  ex->x3= x3 - 2*PIXEL;
  ex->x4= x4 + 2*PIXEL;
  ex->y1= y1;
  ex->y2= y2;
  ex->y3= y3 - 2*PIXEL;
  ex->y4= y4 + 2*PIXEL;
}

void
tex_font_rep::draw (ps_device dev, string s, SI ox, SI y) {
  register int i;
  switch (status) {
  case TEX_ANY:
    break;
  case TEX_CM:
    for (i=0; i<N(s); i++)
      if ((s[i] & 128) != 0) {
	accented_draw (dev, s, ox, y);
	return;
      }
    break;
  case TEX_EC:
  case TEX_LA:
  case TEX_ADOBE:
    for (i=0; i<N(s); i++)
      if (s[i]=='<') {
	special_draw (dev, s, ox, y);
	return;
      }
    break;
  }

  SI  x= ox;
  int n= N(s);
  int m= n+16;
  SI str[n], buf[m], ker[m];
  for (i=0; i<n; i++) str[i]= ((QN) s[i]);
  tfm->execute (str, n, buf, ker, m);

  for (i=0; i<m; i++) {
    register int c= buf[i];
    bitmap_char bmc= pk->get (c);
    if (nil (bmc)) continue;
    dev->draw (c, pk, x, y);
    x += conv (tfm->w(c)+ ker[i]);
  }
}

bitmap_char
tex_font_rep::get_bitmap (string s) {
  register int i;
  switch (status) {
  case TEX_ANY:
    break;
  case TEX_CM:
    for (i=0; i<N(s); i++)
      if ((s[i] & 128) != 0)
	return font_rep::get_bitmap (s);
    break;
  case TEX_EC:
  case TEX_LA:
    if (s == "<less>") s= "<";
    if (s == "<gtr>") s= ">";
    break;
  }
  if (N(s)!=1) return font_rep::get_bitmap (s);
  int c= ((QN) s[0]);
  bitmap_char bmc= pk->get (c);
  if (nil (bmc)) return font_rep::get_bitmap (s);
  return bmc;
}

/******************************************************************************
* Handle <, > and (in the future?) other special characters
******************************************************************************/

void
tex_font_rep::special_get_extents (string s, text_extents& ex) {
  register int i, j;
  for (i=0; i<N(s); i++)
    if (s[i]=='<') break;
  get_extents (s (0, i), ex);
  for (j=i+1; j<N(s); j++)
    if (s[j]=='>') break;
  if (j<N(s)) j++;

  SI x;
  text_extents ey;
  int temp= status;
  status= TEX_ANY;
  string r= s (i, j);
  if (r == "<less>") r= "<";
  else if (r == "<gtr>") r= ">";
  get_extents (r, ey);
  x= ex->x2;
  ex->x1= min (ex->x1, x+ ey->x1); ex->y1= min (ex->y1, ey->y1);
  ex->x2= max (ex->x2, x+ ey->x2); ex->y2= max (ex->y2, ey->y2);
  ex->x3= min (ex->x3, x+ ey->x3); ex->y3= min (ex->y3, ey->y3);
  ex->x4= max (ex->x4, x+ ey->x4); ex->y4= max (ex->y4, ey->y4);
  status= temp;
  
  get_extents (s (j, N(s)), ey);
  x= ex->x2;
  ex->x1= min (ex->x1, x+ ey->x1); ex->y1= min (ex->y1, ey->y1);
  ex->x2= max (ex->x2, x+ ey->x2); ex->y2= max (ex->y2, ey->y2);
  ex->x3= min (ex->x3, x+ ey->x3); ex->y3= min (ex->y3, ey->y3);
  ex->x4= max (ex->x4, x+ ey->x4); ex->y4= max (ex->y4, ey->y4);
}

void
tex_font_rep::special_draw (ps_device dev, string s, SI x, SI y) {
  register int i, j;
  text_extents ex;
  for (i=0; i<N(s); i++)
    if (s[i]=='<') break;
  draw (dev, s (0, i), x, y);
  get_extents (s (0, i), ex);
  x += ex->x2;
  for (j=i+1; j<N(s); j++)
    if (s[j]=='>') break;
  if (j<N(s)) j++;

  int temp= status;
  status= TEX_ANY;
  string r= s (i, j);
  color c= dev->get_color ();
  if (r == "<less>") r= "<";
  else if (r == "<gtr>") r= ">";
  else dev->set_color (dev->red);
  draw (dev, r, x, y);
  dev->set_color (c);
  get_extents (r, ex);
  x += ex->x2;
  status= temp;
  
  draw (dev, s (j, N(s)), x, y);
}

/******************************************************************************
* Handle accents
******************************************************************************/

static char Cork_unaccented[128]= {
  'A', ' ', 'C', 'C', 'D', 'E', ' ', 'G',
  'L', 'L', ' ', 'N', 'N', ' ', 'O', 'R',
  'R', 'S', 'S', 'S', 'T', 'T', 'U', 'U',
  'Y', 'Z', 'Z', 'Z', ' ', 'I', 'd', ' ',
  'a', ' ', 'c', 'c', 'd', 'e', ' ', 'g',
  'l', 'l', ' ', 'n', 'n', ' ', 'o', 'r',
  'r', 's', 's', 's', 't', 't', 'u', 'u',
  'y', 'z', 'z', 'z', ' ', 60 , 62 , ' ',
  'A', 'A', 'A', 'A', 'A', 'A', 29 , 'C',
  'E', 'E', 'E', 'E', 'I', 'I', 'I', 'I',
  'D', 'N', 'O', 'O', 'O', 'O', 'O', 30 ,
  31 , 'U', 'U', 'U', 'U', 'Y', ' ', ' ',
  'a', 'a', 'a', 'a', 'a', 'a', 26 , 'c',
  'e', 'e', 'e', 'e', 16 , 16 , 16 , 16 ,
  'd', 'n', 'o', 'o', 'o', 'o', 'o', 27 ,
  28 , 'u', 'u', 'u', 'u', 'y', ' ', 25
};

static char Cork_accents[128]= {
  21 , ' ', 19 , 20 , 20 , 20 , ' ', 21 ,
  19 , 39 , ' ', 19 , 20 , ' ', 125, 19 ,
  20 , 19 , 20 , 24 , 20 , 24 , 125, 23 ,
  127, 19 , 20 , 95 , ' ', 95 , 22 , ' ',
  21 , ' ', 19 , 20 , 20 , 20 , ' ', 21 ,
  19 , 39 , ' ', 19 , 20 , ' ', 125, 19 ,
  20 , 19 , 20 , 24 , 20 , 24 , 125, 23 ,
  127, 19 , 20 , 95 , ' ', ' ', ' ', ' ',
  18 , 19 , 94 , 126, 127, 23 , ' ', 24 ,
  18 , 19 , 94 , 127, 18 , 19 , 94 , 127,
  22 , 126, 18 , 19 , 94 , 126, 127, ' ',
  ' ', 18 , 19 , 94 , 127, 19 , ' ', ' ',
  18 , 19 , 94 , 126, 127, 23 , ' ', 24 ,
  18 , 19 , 94 , 127, 18 , 19 , 94 , 127,
  22 , 126, 18 , 19 , 94 , 126, 127, ' ',
  ' ', 18 , 19 , 94 , 127, 19 , ' ', ' '
};

static string
get_unaccented (string s) {
  int i;
  string r(N(s));
  for (i=0; i<N(s); i++)
    if ((s[i] & 128) == 0) r[i]= s[i];
    else {
      char c= Cork_unaccented[s[i] & 127];
      if (c==' ') r[i]= s[i];
      else r[i]= Cork_unaccented[s[i] & 127];
    }
  return r;
}

static string
get_accents (string s) {
  int i;
  string r(N(s));
  for (i=0; i<N(s); i++)
    if ((s[i] & 128) == 0) r[i]= ' ';
    else r[i]= Cork_accents[s[i] & 127];
  return r;
}

void
tex_font_rep::accented_get_extents (string s, text_extents& ex) {
  status= TEX_ANY;
  register int i;
  string acc= get_accents (s);
  s= get_unaccented (s);
  get_extents (s, ex);

  for (i=0; i<N(acc); i++)
    if (acc[i] != ' ') {
      SI xx, yy;
      text_extents ey, ez;
      get_extents (s(0,i+1), ey); xx= ey->x2;
      get_extents (s  [i]  , ey);
      get_extents (acc[i]  , ez);
      xx -= (((ey->x2 - ey->x1) + (ez->x2 - ez->x1)) >> 1);
      yy  = ey->y4- ysup;
      if (acc[i] == 24) yy= PIXEL;
      else xx += (SI) (((double) yy) * slope);
      ex->x3 = min (ex->x3, xx + ez->x3);
      ex->y3 = min (ex->y3, yy + ez->y3);
      ex->x4 = max (ex->x4, xx + ez->x4);
      ex->y4 = max (ex->y4, yy + ez->y4);
    }

  status= TEX_CM;
}

void
tex_font_rep::accented_draw (ps_device dev, string s, SI x, SI y) {
  status= TEX_ANY;
  register int i;
  string acc= get_accents (s);
  s= get_unaccented (s);
  draw (dev, s, x, y);

  for (i=0; i<N(acc); i++)
    if (acc[i] != ' ') {
      SI xx, yy;
      text_extents ey, ez;
      get_extents (s(0,i+1), ey); xx= ey->x2;
      get_extents (s  [i]  , ey);
      get_extents (acc[i]  , ez);
      xx -= (((ey->x2 - ey->x1) + (ez->x2 - ez->x1)) >> 1);
      yy  = ey->y4- ysup;
      if (acc[i] == 24) yy= PIXEL;
      else xx += (SI) (((double) yy) * slope);
      dev->draw (acc[i], pk, x+ xx, y+ yy);
    }

  status= TEX_CM;
}

#undef conv(x)

/******************************************************************************
* Interface
******************************************************************************/

font
tex_font (display dis, string family, int size, int dpi, int dsize) {
  string name= "tex:" * family * as_string (size) * "@" * as_string(dpi);
  return make<font> (name,
    new tex_font_rep (dis, name, TEX_ANY, family, size, dpi, dsize));
}

font
tex_cm_font (display dis, string family, int size, int dpi, int dsize) {
  string name= "cm:" * family * as_string (size) * "@" * as_string(dpi);
  return make<font> (name,
    new tex_font_rep (dis, name, TEX_CM, family, size, dpi, dsize));
}

font
tex_ec_font (display dis, string family, int size, int dpi, int dsize) {
  string name= "ec:" * family * as_string (size) * "@" * as_string(dpi);
  return make<font> (name,
    new tex_font_rep (dis, name, TEX_EC, family, size, dpi, dsize));
}

font
tex_la_font (display dis, string family, int size, int dpi, int dsize) {
  string name= "la:" * family * as_string (size) * "@" * as_string(dpi);
  return make<font> (name,
    new tex_font_rep (dis, name, TEX_LA, family, size, dpi, dsize));
}

font
tex_adobe_font (display dis, string family, int size, int dpi, int dsize) {
  string name= "adobe:" * family * as_string (size) * "@" * as_string(dpi);
  return make<font> (name,
    new tex_font_rep (dis, name, TEX_ADOBE, family, size, dpi, dsize));
}

#endmodule // code_tex_font
